"""Submodule containing the IntervalsInfo class."""


class IntervalsInfo:
    """A class representing intervals"""

    def __init__(self):
        """Declaring the properties of the Intervals Info class"""
        self.child_count_per_interval = None
        self.parent_count_per_interval = None
        self.child_cell_weights = None
